//-------------------------  1994 by James G. Stout --------------------------------
// File		: cdefTab.c
// Date		: July 24, 1994
// Author	: Jim Stout
// Purpose	: A CDEF to provide a "Tabbed" panel for use in Tab Dialogs
//			: use procID :   109 * 16 + varCode 
//			:	  when calling NewControl() or in your resource template.
//			:
//			: This control can have a maximum of 5 rows of tabs.  By default,  
//			: there are 4 tabs per row, giving a total of 20 tabs.  The 
//			: control Max field should contain the number of tabs to draw.
//			:
//			: This default can be changed in 2 ways:
//			:
//			: 1. Display a single row of tabs:
//			:
//			:	Variation code 2 will force a single row of tabs, with the
//			:	number of tabs specified in the control Max field.
//			:
//			: 2. Increase or decrease the number of tabs per row:
//			:
//			:	The LoWord of the control refCon can be used to specify from
//			:	2 to 8 tabs per row.  If zero, there will be 4 tabs per row.
//			:
//			: This CDEF supports varCodes of:
//			:
//			:		0 - tab label is Geneva 9, bold for active tab
//			:		1 - tab label is System, underline for active tab
//			:		2 - 1 row of tabs, contrlMax tabs.
//			:		4 - * not used *
//			:		8 - useWindFont for tab label, bold for active tab
//			:
// 			: Pretty straight forward code here, only trick here is to use the 
// 			: contrlMin value as the height of the control.  This is done to keep
// 			: the Control Manager from thinking the control overlays the controls
// 			: you want inside the Panel.  Set the real control rect to some value
// 			: sufficient to see the title in your resource editor and set contrlMin
// 			: to the height of the panel you want drawn.
//			:
// ** IMPORTANT INFORMATION **
//			:
//			: Since the Control Manager doesn't know about the extra drawing done
//			: by this control, some extra handling is required for update events.
//			: Just calling UpdtControl() will not suffice since there could be a
//			: portion of this control that needs to be drawn that is outside the
//			: "official" control rect.
//			:
//			: So, make sure you call Draw1Control() in response to an update for
//			: any control that uses this CDEF.  Calling DrawControls is ok too, but
//			: will draw all of the controls in the window.
//			:
//			: If you find a use for this, I'd love to know about it.  Bug reports
//			: are always interesting.
//			:
//			: Internet    : JimS@WRQ.COM(work hours, PST)
//			: AppleLink   : WRQ			(daily)
//			: CompuServe  : 73240,2052	(weekly or so)
//			: AOL         : JasG		(weekly or so)
//			: eWorld      : Jim Stout	(weekly or so)
//----------------------------------------------------------------------------------

//#define _DEBUGCDEF 1

#include <SetUpA4.h>

#include "grayCDEF.h"
#include "colorCDEF.h"
#include "miscCDEF.h"

#define MAXROWS 5			// maximum number of rows of tabs
#define MINCOLS 2			// minimum number of tabs per row
#define MAXCOLS 8			// maximum number of tabs per row
#define DEFCOLS 4			// default number of tabs per row
//----------------------------------------------------------------------------------
// Variation codes
//----------------------------------------------------------------------------------

#define sysFontTabs		0x0001
#define oneTabRow		0x0002
#define useWFont		0x0008			// draw control with window font

//----------------------------------------------------------------------------------
//	CDEF private data
//----------------------------------------------------------------------------------
typedef struct {
short	txFont;				// font to use for tab labels
short	txSize;				// size to use for tab labels
short	cdefHt;				// true height of tab panel
short	rowWidth;			// width of all tabs on a row
short	rowHt;				// height of one tab
short	prevTab;			// previously active tab
short	nRows;				// number of rows of tabs
short	nCols;				// number of tabs in a row
short	rowOrder[MAXROWS];
}cdefData,**cdefHdl;

//----------------------------------------------------------------------------------
//	Function prototypes
//----------------------------------------------------------------------------------
static void		doInit			(ControlHandle cHdl, short varCode, short txF, short txS);
static long		doTest			(ControlHandle theControl, long param);
static void		doDraw			(ControlHandle cHdl, short varCode);
static void		drawTitle		(ControlHandle cHdl, short varCode, Rect *r, 
									Boolean bgInColor, Boolean rowChg, 
									short part, short prevTab, short lnHt, short lnAdj);
static void		getNextTitle	(Str255 t, Str255 s, short num);
static short	getTabWidth		(ControlHandle cHdl, short row);
static short	getMaxCols		(ControlHandle cHdl, short row);

#ifdef _DEBUGCDEF
pascal long CDmain (short varCode, ControlHandle theControl, short message, long param);
pascal long CDmain (short varCode, ControlHandle theControl, short message, long param)
#else

//==================================================================================
//	CDEF main
//==================================================================================

pascal long main (short varCode, ControlHandle theControl, short message, long param)
#endif
{
	short		txF,txS;
	long		ret = 0L;
	Rect		r;
	Point		p;
	GrafPtr		thisPort;
	cdefHdl		hCdef;
		
	RememberA0();
	SetUpA4();
		
	if((**theControl).contrlData)
		HLock((**theControl).contrlData);
				
	GetPort(&thisPort);
	
//----------------------------------------------------------------------------------
// set window font & size info	
//----------------------------------------------------------------------------------
	
	txF = thisPort->txFont;
	txS = thisPort->txSize;
	if(varCode & sysFontTabs) {					// use system font	
		thisPort->txFont = SysFontFam;
		thisPort->txSize = SysFontSize;
	}
	else
	if(!(varCode & useWFont)) {					// default to geneva 9
		TextFont(geneva);
		TextSize(9);
	}
	
//----------------------------------------------------------------------------------
// Process messages to control	
//----------------------------------------------------------------------------------

	switch(message) {
		case initCntl:
			doInit(theControl, varCode, txF, txS);
		break;
		case dispCntl:
			if((**theControl).contrlData) {
				HUnlock((**theControl).contrlData);
				DisposeHandle((**theControl).contrlData);
				(**theControl).contrlData = 0;
			}
		break;
		case drawCntl:
			if ((**theControl).contrlVis != 0 &&
				((WindowPeek)(**theControl).contrlOwner)->visible) {
				doDraw(theControl, varCode);
			}
		break;
		case testCntl:
			p.v=HiWord(param);
			p.h=LoWord(param);
			if(PtInRect(p,&(*theControl)->contrlRect)) {
				ret = doTest(theControl, param);
			}
		break;
		case calcCRgns:
			RectRgn((RgnHandle)(param & 0x7fffffffL), &(*theControl)->contrlRect);
		break;
		case calcCntlRgn:
		case calcThumbRgn:
			RectRgn((RgnHandle)(param), &(*theControl)->contrlRect);
		break;
	}

//----------------------------------------------------------------------------------
// restore window font & size info	
//----------------------------------------------------------------------------------

	if(!(varCode & useWindFont)) {
		thisPort->txFont = txF;
		thisPort->txSize = txS;
	}
	if((**theControl).contrlData)
		HUnlock((**theControl).contrlData);
	RestoreA4();
	return (ret);
}
//==================================================================================
// Initialize our private control data
//==================================================================================
static void doInit(ControlHandle theControl, short varCode, short txF, short txS)
{
	cdefHdl		hCdef;
	Rect		r;
	short		wid,ht,notch,rows,maxTabs,maxCols;
	
	hCdef = (cdefHdl)NewHandle(sizeof(cdefData));
	if(hCdef) {
//----------------------------------------------------------------------------------
// initialize row order array - this contains the 'order' of the rows which can
// change in multi-row tab panels.
//----------------------------------------------------------------------------------
			
		for(rows=1;rows<MAXROWS;rows++)
			(**hCdef).rowOrder[rows-1] = rows;

//----------------------------------------------------------------------------------
// bounds check on number of columns, rows, control maximum & control value
//----------------------------------------------------------------------------------
		if(varCode & oneTabRow) {
			maxCols = (**theControl).contrlMax;
		}
		else {
			maxCols = LoWord((**theControl).contrlRfCon);	// number of columns
			if(maxCols < MINCOLS || maxCols > MAXCOLS)		// default to 4
				maxCols = DEFCOLS;
		}
			
		maxTabs = MAXROWS*maxCols;
	
		if((**theControl).contrlMax <= 0)					// should be the total
			(**theControl).contrlMax = maxCols;				// number of tabs
		else
		if((**theControl).contrlMax > maxTabs)
			(**theControl).contrlMax = maxTabs;
			
		if((**theControl).contrlValue > maxTabs)			// the 'active' tab
			(**theControl).contrlValue = 1;
			
		rows = (**theControl).contrlMax / maxCols;
		if((**theControl).contrlMax % maxCols)				// odd number of tabs,
			rows++;											// so add a row

//----------------------------------------------------------------------------------
// various size limits
//----------------------------------------------------------------------------------
		
		r = (**theControl).contrlRect;
		wid = r.right-r.left;
		ht = r.bottom-r.top;
//----------------------------------------------------------------------------------		
// the "notch" in the right corner must be no more than half the control width
//----------------------------------------------------------------------------------

		notch = HiWord((**theControl).contrlRfCon);
		if(notch < 0 || notch > wid/2)
			notch = 0;
//----------------------------------------------------------------------------------		
// width of a row of tabs must be adjusted for 'notch' and a 10 pixel inset
// from the right edge of control for each row.
//----------------------------------------------------------------------------------
	
		(**hCdef).rowWidth = wid - notch - ((rows-1)*10);

//----------------------------------------------------------------------------------
// The height of the control is really in the contrlMin field, not contrlRect
//----------------------------------------------------------------------------------
		
		(**hCdef).cdefHt = (**theControl).contrlMin;
		if((**hCdef).cdefHt < ht)
			(**hCdef).cdefHt = 2*ht;
//----------------------------------------------------------------------------------
// calcuate the height of one row of tabs
//----------------------------------------------------------------------------------
		(**hCdef).rowHt = ht / rows;
		
//----------------------------------------------------------------------------------
// set rows & columns
//----------------------------------------------------------------------------------

		if(varCode & oneTabRow) {					// only want 1 row of tabs
			(**hCdef).nRows = 1;
			(**hCdef).nCols = (**theControl).contrlMax;
		}
		else {
			(**hCdef).nRows = rows;
			(**hCdef).nCols = maxCols;
		}
		
		(**theControl).contrlMin = 1;
		(**hCdef).prevTab = 0;
		(**hCdef).txFont = txF;
		(**hCdef).txSize = txS;
		
		(**theControl).contrlData = (Handle)hCdef;
		HLock((**theControl).contrlData);
	}
}
//==================================================================================
// Check for a mouse down in one of our tabs - tab information is stored in the
// rowOrder.
//==================================================================================
static long doTest(ControlHandle theControl, long param)
{
	short		item,row1,row,col,maxRow,maxCol,tabCnt,tabHt,tabWid;
	Rect		r,r2;
	Point		p;
	cdefHdl		hCdef;
	
	hCdef = (cdefHdl)(**theControl).contrlData;
	if(!hCdef)
		return;
	
	p.v=HiWord(param);
	p.h=LoWord(param);
		
	tabHt = (**hCdef).rowHt;
	maxRow = (**hCdef).nRows;
	tabCnt = (**theControl).contrlMax;
	r2 = r = (**theControl).contrlRect;
	
	for(row=1;row<=maxRow;row++) {
		row1 = (**hCdef).rowOrder[row-1];
		tabWid = getTabWidth(theControl, row1);
		maxCol = getMaxCols(theControl, row1);
		r2.right = r2.left+tabWid;
		r2.top = r2.bottom-tabHt;
		r = r2;
		for(col=1;col<=maxCol;col++) {
			item = (row1-1)*(**hCdef).nCols + col;
			if(item > tabCnt)
				break;
			if(PtInRect(p, &r)) {
				return((long)item);
			}
			OffsetRect(&r,tabWid-1, 0);
		}
		OffsetRect(&r2, 10, -tabHt);
	}
	return (0L);
}

//==================================================================================
// draw the tab panel.  Lots and lots of tedious line drawing - all done onscreen,
// no real need for offscreen drawing.
//==================================================================================
static void doDraw(ControlHandle theControl, short varCode)
{
	short			tabWidth,tabHt,notch,rightEdge,bottomEdge;
	short			inx,row,maxRow,newRow,col,maxCol,delta,pLeft;
	short			tabCnt,activeTab,currTab,prevTab;
	short			lnHt,lnAdj;
	Rect			panel,tab,fullRect,edgeRect,allTabs;
	FontInfo		f;
	RGBColor		rgbWhite = {-1,-1,-1};
	RGBColor		rgbGray,saveFore,saveBack;
	Boolean			inColor = false, bgInColor = false;
	Boolean			rowChg=false, partialRow = false;
	cdefHdl			hCdef;
		
	hCdef = (cdefHdl)(**theControl).contrlData;
	if(!hCdef)
		return;
	
//----------------------------------------------------------------------------------
// if called to Hilite the control, just set the control value and leave
// because we don't hilite.
//----------------------------------------------------------------------------------
	
	if((**theControl).contrlHilite) {
		(**theControl).contrlValue = (**theControl).contrlHilite;
		return;
	}

//----------------------------------------------------------------------------------
// called to draw the control
// adjust the rowOrder if needed (user clicked on row other than lowest)
//----------------------------------------------------------------------------------
	
	newRow = ((**theControl).contrlValue-1) / (**hCdef).nCols + 1;
			
	if(newRow != (**hCdef).rowOrder[0]) {
		rowChg = true;
		(**hCdef).rowOrder[0] = newRow;
	
		maxRow = (**hCdef).nRows;
		row = 0;
		for(inx=1;inx<=maxRow-1;inx++) {
			row++;
			if(row == newRow)
				row++;
			(**hCdef).rowOrder[inx] = row;
		}
	}

//----------------------------------------------------------------------------------
// initialize for drawing
//----------------------------------------------------------------------------------
	
	prevTab = (**hCdef).prevTab;
	activeTab = (**theControl).contrlValue;
	tabCnt = (**theControl).contrlMax;
	tabHt = (**hCdef).rowHt;
	maxRow = (**hCdef).nRows;
	
	if(maxRow != (tabCnt / (**hCdef).nCols))			// last row partial?
		partialRow = true;

//----------------------------------------------------------------------------------
// get the font info for drawing title
//----------------------------------------------------------------------------------
	
	GetFontInfo(&f);
	lnAdj = f.descent + f.leading;				
	lnHt = f.ascent + lnAdj;

//----------------------------------------------------------------------------------
// get the true Rect for the control
//----------------------------------------------------------------------------------

	panel = tab = (**theControl).contrlRect;
	panel.bottom = panel.top + (**hCdef).cdefHt;
	fullRect = panel;
	panel.right = panel.right-((maxRow-1) * 10);
	
//----------------------------------------------------------------------------------
// get color information and set flags for color/3D drawing
//----------------------------------------------------------------------------------
	
	if(getPixDepth(&fullRect) > 2) {
		saveColors(&saveFore, &saveBack);
		setPartColor(theControl, cFrameColor, true);
		getGray(&rgbGray);								// used later on
		inColor = true;
		bgInColor = true;
		if(saveBack.red == 65535 &&						// is bg white?
			saveBack.green == 65535 &&
			saveBack.blue == 65535)
			bgInColor = false;
	}			
	
//----------------------------------------------------------------------------------
// Draw each tab by looping thru rows (by rowOrder[]) & columns
//----------------------------------------------------------------------------------

	for(inx=1;inx<=maxRow;inx++) {
		row = (**hCdef).rowOrder[inx-1];
		
		tabWidth = getTabWidth(theControl, row);		// may be different
		tab.right = tab.left+tabWidth;					// for some rows
		tab.top = tab.bottom - tabHt;
	
		bottomEdge = tab.bottom;
		rightEdge = panel.left + (**hCdef).rowWidth;
		notch = panel.right - rightEdge;
		
		if(rowChg && partialRow) {						// clear old tabs
			allTabs = tab;
			allTabs.right = rightEdge-4;
			allTabs.left+=4;
			EraseRect(&allTabs);
		}
		
		maxCol = getMaxCols(theControl, row);
		
		for(col=1;col<=maxCol;col++) {
			currTab = (**hCdef).nCols * (row-1) + col;
			if(currTab > tabCnt)
				break;
			if(col == maxCol)							// last tab must 
				tab.right = rightEdge;					// fill control rect
	
//----------------------------------------------------------------------------------
// draw the tab title
//----------------------------------------------------------------------------------
			
			if(inColor)
				setPartColor(theControl, cTextColor, true);
			drawTitle (theControl, varCode, &tab, bgInColor, rowChg,
						currTab, prevTab, lnHt, lnAdj);
	
//----------------------------------------------------------------------------------
// draw shadow inside tab frame
//----------------------------------------------------------------------------------
			
			if(bgInColor) {
			
//----------------------------------------------------------------------------------
// remove wide shadow if this is not the active tab
//----------------------------------------------------------------------------------

				if(currTab != activeTab) {
					RGBForeColor(&saveBack);
					MoveTo(tab.left+2,tab.bottom-1);
					LineTo(tab.left+2,tab.top+3);			// left edge
					MoveTo(tab.left+3,tab.top+2);
					LineTo(tab.right-4,tab.top+2);			// top edge
					if(inx == 1) {
						MoveTo(tab.right-3, tab.top+3);		// right corner
						LineTo(tab.right-3, tab.bottom-1);	// right edge
					}
				}
				
//----------------------------------------------------------------------------------
// draw white shadow along left & top edges of tab
//----------------------------------------------------------------------------------

				RGBForeColor(&rgbWhite);
				MoveTo(tab.left+1,tab.bottom-1);
				if(currTab == activeTab)
					Move(0,1);
				LineTo(tab.left+1,tab.top+3);				// left edge
				LineTo(tab.left+3,tab.top+1);				// left corner
				LineTo(tab.right-5,tab.top+1);				// top edge
				if(currTab == activeTab) {
					Move(1,1);
					LineTo(tab.left+2, tab.top+2);
					LineTo(tab.left+2, tab.bottom+1);
				}

//----------------------------------------------------------------------------------
// draw or remove white shadow from top edge of first panel
// lots of work to avoid flicker...
//----------------------------------------------------------------------------------
				if(inx == 1) {
					if(currTab != activeTab) {
						RGBForeColor(&rgbWhite);
						MoveTo(tab.left,tab.bottom+1);
						if(col == 1)
							Move(3,0);
						else
							Move(-1,0);
						if(col == maxCol) {
							LineTo(tab.right-3,tab.bottom+1);
							Move(-1,0);
						}
						else
							LineTo(tab.right,tab.bottom+1);
						Move(0,1);
						if(col == 1)
							LineTo(tab.left+3,tab.bottom+2);
						else
							LineTo(tab.left-2,tab.bottom+2);
					}
					else {
						RGBForeColor(&saveBack);
						MoveTo(tab.left+2,tab.bottom+1);
						if(col == 1 || currTab == activeTab)
							Move(1,0);
						if(col == maxCol) {
							LineTo(tab.right-3,tab.bottom+1);
							Move(-1,0);
						}
						else
							LineTo(tab.right,tab.bottom+1);
						Move(0,1);
						if(col == 1)
							LineTo(tab.left+3,tab.bottom+2);
						else
							LineTo(tab.left+2,tab.bottom+2);
					}
				}

//----------------------------------------------------------------------------------
// draw gray shadow along right edge of tab
//----------------------------------------------------------------------------------
	
				RGBForeColor(&rgbGray);
				MoveTo(tab.right-4, tab.top+1);
				LineTo(tab.right-2, tab.top+3);				// right corner
				if(currTab == activeTab) {
					LineTo(tab.right-2, tab.bottom);		// right edge
					Move(-1,1);
					LineTo(tab.right-3, tab.top+2);
				}
				else
					LineTo(tab.right-2, tab.bottom-1);
					if(col == maxCol && inx != 1)
						Line(0,2);
			}
			
//----------------------------------------------------------------------------------
// draw the tab frame
//----------------------------------------------------------------------------------
				
			if(inColor)
				setPartColor(theControl, cFrameColor, true);
			MoveTo(tab.left,tab.bottom);			
			LineTo(tab.left,tab.top+3);					// left edge
			LineTo(tab.left+3,tab.top);					// left corner
			LineTo(tab.right-4,tab.top);				// top edge
			LineTo(tab.right-1,tab.top+3);				// right corner
			LineTo(tab.right-1,tab.bottom);				// right edge
			
			MoveTo(tab.right-2,tab.bottom);
			if(currTab != activeTab && inx == 1)		// bottom edge needed
				LineTo(tab.left,tab.bottom);
			else 
			if(inx == 1) {								// bottom edge not needed
				edgeRect = tab;
				edgeRect.left++;
				edgeRect.right--;
				edgeRect.top = edgeRect.bottom++;
				if(inColor) {							// erase to bg color
					RGBForeColor(&saveBack);
					if(bgInColor) {
						edgeRect.left+=2;
						edgeRect.right-=2;
					}
				}
				EraseRect(&edgeRect);
			}
			OffsetRect(&tab,tabWidth-1, 0);				// next tab to draw
		}

//----------------------------------------------------------------------------------
// draw panel outline
//----------------------------------------------------------------------------------
			
		if(inColor)
			setPartColor(theControl, cFrameColor, true);
	
		if(inx == 1) {									// first panel
			pLeft = panel.left;
			MoveTo(pLeft,bottomEdge);
			LineTo(pLeft,panel.bottom-1);				// left edge
		}
		else
			pLeft = panel.right-10;
		MoveTo(pLeft,panel.bottom-1);	
		LineTo(panel.right-1,panel.bottom-1);			// bottom edge
		LineTo(panel.right-1,bottomEdge);				// right edge
	
		if(notch) {										// top edge (if needed)
			Line(-notch,0);
		}

//----------------------------------------------------------------------------------
// draw panel shadowing inside panel frame
//----------------------------------------------------------------------------------
	
		if(bgInColor) {

//----------------------------------------------------------------------------------
// draw white shadow along left edge of panel
//----------------------------------------------------------------------------------

			RGBForeColor(&rgbWhite);
			PenSize(2,2);
			if(inx == 1) {
				MoveTo(pLeft+1,bottomEdge+1);
				LineTo(pLeft+1,panel.bottom-3);
			}
			PenSize(1,1);

//----------------------------------------------------------------------------------
// draw white shadow along top edge of panel
//----------------------------------------------------------------------------------
	
			if(notch) {
				MoveTo(panel.right-3,bottomEdge+1);
				Line(-notch+1,0);
				if(inx == 1) {
					Move(-1,1);
					Line(notch-1,0);
				}
			}

//----------------------------------------------------------------------------------
// draw gray shadow along bottom edge of panel
//----------------------------------------------------------------------------------
	
			RGBForeColor(&rgbGray);
			PenSize(1,1);
			MoveTo(panel.right-3, panel.bottom-3);
			if(inx == 1) {								// first panel
				LineTo(pLeft+3,panel.bottom-3);
				Move(-1,1);
			}
			else
				MoveTo(pLeft,panel.bottom-2);
			LineTo(panel.right-2, panel.bottom-2);

//----------------------------------------------------------------------------------
// draw gray shadow along right edge of panel
//----------------------------------------------------------------------------------
			
			if(inx == 1 || notch) {						// first panel
				LineTo(panel.right-2, bottomEdge+1);
				Move(-1,1);
			}
			else {
				LineTo(panel.right-2, tab.top+3);
				Move(-1,0);
			}
			if(inx == 1)
				LineTo(panel.right-3, panel.bottom-2);
		}

//----------------------------------------------------------------------------------
// prepare to draw the next panel
//----------------------------------------------------------------------------------

		OffsetRect(&panel, 10, 0);						// draw next panel, which
		panel.bottom-=10;								// is shifted up & right
		OffsetRect(&tab, 0, -tabHt);					// shift tab up to next row
		tab.left = panel.left;
	}
	
	if(inColor)
		restoreColors(&saveFore, &saveBack);
		
	(**hCdef).prevTab = (**theControl).contrlValue;
}

//==================================================================================
// draw the titles for the tabs.  Must take care of redrawing a previous tab in
// a normal face and active tab in a bold face.
//==================================================================================
static void drawTitle (ControlHandle theControl, short varCode, Rect *r,  
						Boolean bgInColor, Boolean rowChg, 
						short part, short prevTab, 
						short lnHt, short lnAdj)
{
	short 		lnCnt=1,lnSpace,ctlSpace,v,h;
	Str255		s;
	Rect		tRect;
	RgnHandle	oldClip;
	RGBColor	saveFore, rgbWhite = {-1, -1, -1};

//----------------------------------------------------------------------------------
// clear title area if needed - click on new tab or we are swapping rows around
//----------------------------------------------------------------------------------

	if(part == prevTab || part == (**theControl).contrlValue || rowChg) {
		tRect = *r;
		InsetRect(&tRect, 3, 3);
		tRect.bottom+=3;
		EraseRect(&tRect);
	}
	
	getNextTitle((*theControl)->contrlTitle, s, part);
	if(s[0]) {								// is there a title?
		if(bgInColor)
			GetForeColor(&saveFore);
		oldClip = NewRgn();
		GetClip(oldClip);
	
		if(part == (**theControl).contrlValue) {
			if(varCode & sysFontTabs)
				TextFace(underline);
			else
				TextFace(bold);
		}
		v = (*r).bottom;					// baseline
		ctlSpace = (*r).bottom - (*r).top;	// available height
		v-=(ctlSpace - lnHt)/2;				// minus free space
		v-=lnAdj;							// minus descent+leading

		h = (*r).left + ((*r).right - (*r).left - StringWidth(s))/2;

		ClipRect(r);
		
		if(bgInColor) {						// draw 3D title
			RGBForeColor(&rgbWhite);
			MoveTo(h+1, v+1);
			DrawString(s);
			RGBForeColor(&saveFore);
		}
		MoveTo(h, v);
		DrawString(s);
		
		TextFace(normal);					// cleanup & leave
		SetClip(oldClip);
		DisposeRgn(oldClip);
	}
}
//==================================================================================
// Grab the next tab title from the control title (assumes lines separated by CR)
//==================================================================================
static void getNextTitle(Str255 t, Str255 s, short part)
{
	short	inx=1,len=0,tCnt=1;
	
	s[0] = 0;								// always default to null string
	
	do {
		if(t[inx] == 0x0d)					// next title
			tCnt++;
		else
		if(tCnt == part) {					// return this title
			len++;
			s[0] = len;
			s[len] = t[inx];
		}
		inx++;
	}while (inx <= t[0]);
}
//==================================================================================
// calculate the maximum number of columns in this row of tabs
//==================================================================================

static short getMaxCols (ControlHandle theControl, short row)
{
	short		tabCnt;
	cdefHdl		hCdef;
		
	hCdef = (cdefHdl)(**theControl).contrlData;
	if(!hCdef)
		return;
	
	if(row == (**hCdef).nRows)				// last row
		tabCnt = (**theControl).contrlMax - ((**hCdef).nCols * (row -1));
	else
		tabCnt = (**hCdef).nCols;
	return(tabCnt);
}
//==================================================================================
// calculate the width of a single tab - can be different for last row if it is not
// a complete row.
//==================================================================================

static short getTabWidth(ControlHandle theControl, short row)
{
	short		wid,tabWidth,temp,tabCnt;
	cdefHdl		hCdef;
		
	hCdef = (cdefHdl)(**theControl).contrlData;
	if(!hCdef)
		return;
	
	if(row == (**hCdef).nRows)				// last row
		tabCnt = (**theControl).contrlMax - ((**hCdef).nCols * (row -1));
	else
		tabCnt = (**hCdef).nCols;
	
	
	wid = (**hCdef).rowWidth;			
	
	tabWidth = wid/tabCnt;					// width of 1 tab
	
	temp = wid - (tabWidth-1)*tabCnt;		// round it up a bit
	tabWidth = tabWidth + temp/tabCnt;
	
	return(tabWidth);
}