/*	pipe.h		Pipe handling
 *
 *	Released MacDOS 1.0.1a on 94/01/31
 *	94/03/31	File created
 *	94/05/27	First version completed
 *	Released MacDOS 2.0.02 on 94/06/07
 *	94/06/13	Added pipeParmSeparator
 *				Removed pipeBadParmErr and renumbered the remaining error codes
 *
 *	Note that some of the definitions contained in this file are here because they are shared
 *	by MacDOS and pipe.c, not because you are expected to do something with them when you write
 *	a filter. To avoid confusion, these definitions are grouped at the end of the file.
 *	Unless you intend to modify the pipe mechanism itself (GULP!) ignore them.
 *
 *	A filter process has a structure as follows:
 *		PipeInit(...)
 *		forewer {
 *			PipePoll(...)
 *			perform the filtering function on the buffer returned by PipePoll
 *			if an error occurs while filtering
 *				report the error with PipeReportError(...)
 *			else
 *				forward the result to the next filter with PipeSendData(...)
 *			}
 *
 *--------------------------------------------------------------------------------------------------
 *  Copyright  1994 by Rainbow Hill Pty Ltd.
 *
 *	This program is free software; you can redistribute it and/or modify it under the terms of
 *	the GNU General Public License as published by the Free Software Foundation; either version 2
 *	of the License, or any later version.
 *
 *	This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 *	without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 *	See the GNU General Public License for more details.
 *
 *	You should have received a copy of the GNU General Public License along with this program;
 *	if not, write to the Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */
#ifndef __PIPE__
#define __PIPE__

#include <Errors.h>
#include <PPCToolBox.h>

#define pipeSize			256			/* messages never exceed this length */
#define pipeParmSeparator	'\0'		/* used to separate parameters in the parameter list */
#define pipeFilteringErr	-950		/* OSType extension: to report your filtering errors */

/*
 *	Type definition of the call-back function used to process your filter parameters.
 *	You install this function when you initialise the pipes with PipeInit and the function
 *	PipeGetFromMaster executes it when it receives your filter parameters from the incoming
 *	pipe. Note that MacDOS sends a message of parameters even if it is empty. If your filter
 *	does not handle any parameter, pass nil to PipeInit. PipeGetFromMaster will then ignore
 *	the parameters message sent by MacDOS.
 */
typedef OSErr pipeParmsFun_t(unsigned char *);

/*
 *	PipeInit prepares the filter to receive and send pipe messages. It opens the PPC port and
 *	only returns after establishing a session with the process upstream. It must be called before
 *	any other pipe function and can only be called once.
 *
 *	parmsFun is a pointer to a function which accepts configuration parameters as they appear in
 *	the command line and saves them in variables accessible during filtering. Set it to nil if
 *	your filter is not configurable.
 */
extern void PipeInit(pipeParmsFun_t *parmsFun);

/*
 *	PipePoll handles both pipes and only returns when it receives a message to be filtered or
 *	detects an error condition. That is, it automatically handles all messages unless they are
 *	data messages received from the incoming pipe and destined to the filter itself.
 *
 *	inMess is the pointer to the buffer where PipePoll can store an incoming message. inMess should
 *	be at least pipeSize byte long and will contain a P-string when PipePoll returns successfully.
 */
extern void PipePoll(unsigned char *inMess);

/*
 *	PipeReportError sends an error message to MacDOS. It should be used to report system errors
 *	returned by the Mac toolbox functions while performing the filtering.
 *
 *	err is the error code. To have MacDOS display an error message which does not correspond to any
 *	system error code, set err to noErr. MacDOS will still display errInClear.
 *
 *	errInClear is a proprietary error message that you would like MacDOS to display in the console
 *	window. When you use PipeReportError to report a system error rather than a proprietary
 *	error, set errInClear to nil. MacDOS will then handle err in the standard way.
 */
extern void PipeReportError(OSErr err, unsigned char *errInClear);

/*
 *	PipeSendData sends a message to next filter downstream.
 *
 *	outMess is the pointer to the buffer containing the message to be sent. outMess should
 *	contain a P-string should not occupy more than pipeSize bytes.
 */
extern void PipeSendData(unsigned char *outMess);


/**************************************************************************************************
 *	As explained in the file header, here are definitions which are only exported when compiling  *
 *	MacDOS to ensure consistency between MacDOS and pipe.c . You can safely ignore them.          *
 **************************************************************************************************/

/*-------------------------------------------------------------------------- Constant definitions */

#define pipeNoSession		0L			/* because PPC sessions numbers are never zero */
#define pipeTimeoutDefault	2			/* default for the timeout in seconds (for MacDOS only) */
#define pipeMaxMessDefault	0			/* default for the transmit window size */
#define pipeType			"\ppipe"	/* port type used to implement pipes */
#define pipeScriptStrNum	128			/* 'STR ' containing the script code (msRoman, ...) */
#define pipeMacDOSID		0			/* filter ID of MacDOS itself */

/*
 *	Proprietary error definitions to extend the standard OSErr type.
 *	At the time of writing, <Error.h> does not define codes between -933 and -999.
 *	Use the error code pipeFilteringErr when reporting an error encountered while executing
 *	your filtering code.
 */
enum {
	/* pipeFilteringErr		= -950,	/* to report your filtering errors */

	/* the following errors are used internally within pipe.c and/or MacDOS */
	pipePpcNotDone			= 1,	/* returned in the ioResult field of PPC's Parameter Blocks */
	pipeReturnE				= -951,	/* to indicate reception of a proper data message */
	pipeAnotherDestE		= -952,	/* MacDOS: received message not directed to MacDOS */
	pipeTimeoutE			= -953,	/* MacDOS: requested operation timed out */
	pipeComFilterE			= -954,	/* MacDOS: filter sent a command that only MacDOS should issue */
	pipeSurpriseDataE		= -955,	/* MacDOS: received a data message during initialisation */
	pipeDoNotReportE		= -956,	/* MacDOS: failed silently: an error was already reported */
	pipeSequenceE			= -957,	/* MacDOS: received data message out of sequence */

	/******* insert new errors around this comment line *******/

	pipeUnknownMessErr		= -974,	/* unknown command */
	pipeDoubleInitErr		= -975,	/* you cannot initialise the pipes twice */
	pipeUpstreamErr			= -976,	/* a downstream-only message was received from upstream */
	pipeDownstreamErr		= -977,	/* an upstream-only message was received from downstream */
	pipeFilterRespErr		= -978,	/* an error message was addressed to a filter */
	pipeNoPPCErr			= -979	/* Process-to-process communication not available */
	/* the preceding errors are automatically reported to MacDOS by functions in pipe.c */
	};

/*------------------------------------------------------------------------------ Type definitions */

/*
 *	message types
 */
typedef enum {
	pipeDataMess	= 1,	/* characters to be filtered (only downstream) */
	pipeParmMess	= 2,	/* filter configuration parameters (only downstream) */
	pipeNextMess	= 3,	/* set up a pipe to the next process (only downstream) */
	pipeErrMess		= 4		/* error message for MacDOS (only upstream) */
	} pipeMessType_t;

/*
 *	pipe structure definition
 */
struct pipe_struct {
	long			session;			/* PPC session */
	PPCReadPBRec	readPB;				/* PPC parameter block to read a message */
	unsigned char	readBuff[pipeSize];	/* read buffer */
	PPCWritePBRec	writePB;			/* PPC parameter block to write a message */
	unsigned char	writeBuff[pipeSize];/* write buffer */
	};
typedef struct pipe_struct pipe_t;

/*---------------------------------------------------------------- constant and macro definitions */

/*
 *	PipeIsBreak_M returns a non-zero value if a character event contains a cmd-dot or a cntl-C.
 */
#define PipeIsBreak_M(ev) (														\
	((ev).modifiers & cmdKey) != 0  &&  ((ev).message & charCodeMask) == '.'	\
	||																			\
	((ev).message & charCodeMask) == '\3'										\
	)

/*
 *	The following offsets identify bytes of the userData field.
 */
enum {
	pipeTypeOffs	= 0,	/* message type */
	pipeSrcOffs		= 1,	/* ID of the filter that creates the message */
	pipeDstOffs		= 2,	/* ID of the filter to which the message is directed */
	pipeSeqOffs		= 3		/* source sequence number of the message */
	};

/*
 *	The following two macros encode and extract a character into/from a userData field of
 *	a message.
 */
#define PipeEncUserByte_M(offs, value, destLong) { ((unsigned char *)&(destLong))[offs] = (value); }
#define PipeDecUserByte_M(fromLong, offs)		 (((unsigned char *)&(fromLong))[offs])

/*
 *	PipeEncUserData_M encodes a userData field of a generic message
 */
#define PipeEncUserData_M(destID, messTyp, userData) {			\
	PipeEncUserByte_M(pipeTypeOffs, messTyp, userData);			\
	PipeEncUserByte_M(pipeSrcOffs, pipeFilterID, userData);		\
	PipeEncUserByte_M(pipeDstOffs, destID, userData);			\
	PipeEncUserByte_M(pipeSeqOffs, pipeCurrentSeq, userData);	\
	}

/*
 *	PipeDecUserData_M decodes a userData message field of a message
 */
#define PipeDecUserData_M(userData, typ, src, dst, seq) {	\
	(typ) = PipeDecUserByte_M(userData, pipeTypeOffs);		\
	(src) = PipeDecUserByte_M(userData, pipeSrcOffs);		\
	(dst) = PipeDecUserByte_M(userData, pipeDstOffs);		\
	(seq) = PipeDecUserByte_M(userData, pipeSeqOffs);		\
	}

/*
 *	The following offsets identify bytes of the message buffer.
 */
enum {
	pipeErrMSBOffs		= 1,	/* Most Significant Byte of error code in responses */
	pipeErrLSBOffs		= 2,	/* Least Significant Byte of error code in responses */

	pipeErrMessOffs				/* Offset of the message which can possibly follow the error code */
	};

/*
 *	PipeEncErr_M takes a short integer and encodes it into the two bytes used to store a
 *	value of type OSErr in response messages (Big Endian).
 */
#define PipeEncErr_M(err, buff) {										\
	(buff)[0] = 2;														\
	(buff)[pipeErrMSBOffs] = (unsigned char)(((err) >> 8) & 0x00FF);	\
	(buff)[pipeErrLSBOffs] = (unsigned char)((err) & 0x00FF);			\
	}

/*
 *	PipeDecErr_M takes the two bytes used to store a value of type OSErr in response messages
 *	and converts them to a short integer (Big Endian).
 */
#define PipeDecErr_M(buff) (((short)(buff)[pipeErrMSBOffs] << 8) | (short)(buff)[pipeErrLSBOffs])

/*
 *	PipeEncErrUserData_M encodes the userData and the error code of an error message
 */
#define PipeEncErrUserData_M(errCode, userData, buff) {			\
	PipeEncUserData_M(pipeMacDOSID, pipeErrMess, userData);		\
	PipeEncErr_M(errCode, buff);								\
	}

/*
 *	PipeWaitOnCond_M accepts a condition and returns when the condition ceases to be satisfied
 *	or when the user cancels. res is set to false and it becomes true when the user cancels.
 *	Note that a function would only be possible if we replaced the condition with a call-back
 *	function. The macro works because it operates in the environment of the caller.
 */
#define PipeWaitOnCond_M(cond, res) {							\
	(res) = false;												\
	while ((cond)  &&  (res) == false) (res) = PipeCheckQuit();	\
	}

/*------------------------------------------------------------------------- variable declarations */

extern Str32			pipeName;			/* name of the filter application/process */
extern pipe_t			pipeIn;				/* incoming pipe */
extern pipe_t			pipeOut;			/* outgoing pipe */
extern unsigned char	pipeCurrentSeq;		/* last sequence number received from upstream */
extern unsigned char	pipeFilterID;		/* ID of this filter */

/*------------------------------------------------------------------------- function declarations */

extern OSErr	PipeAcceptInSession(void);
extern Boolean	PipeCheckQuit(void);
extern void		PipeCleanUp(void);
extern OSErr	PipeOpen(void);
extern OSErr	PipePollOnce(pipe_t **pip, long *userData, unsigned char *buff);
extern OSErr	PipeStartWrite(pipe_t *pip, long userData, unsigned char *buff);
extern OSErr	PipeStartNextSession(unsigned char *portName, unsigned char id);

#endif /* __PIPE__ */
